#define __STDC_FORMAT_MACROS
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "evrecord.h"
#include "ev.h"
#include "logread.h"

//sscanf() is a bit slow, so a few dedicated functions for parsing integers
//have been implemented.
static const char *parse_lx(const char *s, unsigned long *value) {
	unsigned long v=0;
	while(*s==' ' || *s=='\t') s++;
	for(;;) {
		char c=*s;
		unsigned long v2;
		if(c>='0' && c<='9')
			v2 = c-'0';
		else if(c>='a' && c<='f')
			v2 = c-'a'+10;
		else if(c>='A' && c<='F')
			v2 = c-'A'+10;
		else
			break;
		v = v*16 + v2;
		s++;
	}
	*value = v;
	return s;
}
static const char *parse_ld(const char *s, long *value) {
	long v=0;
	while(*s==' ' || *s=='\t') s++;
	bool is_negative=false;
	if(*s=='-') { is_negative=true; s++; }
	for(;;) {
		char c=*s;
		if(c>='0' && c<='9')
			v = v*10 + c-'0';
		else
			break;
		s++;
	}
	if(is_negative)
		*value = -v;
	else
		*value = v;
	return s;
}
static const char *parse_lu(const char *s, unsigned long *value) {
	unsigned long v=0;
	while(*s==' ' || *s=='\t') s++;
	for(;;) {
		char c=*s;
		if(c>='0' && c<='9')
			v = v*10 + c-'0';
		else
			break;
		s++;
	}
	*value = v;
	return s;
}
static const char *parse_d(const char *s, int *value) {
	int v=0;
	while(*s==' ' || *s=='\t') s++;
	bool is_negative=false;
	if(*s=='-') { is_negative=true; s++; }
	for(;;) {
		char c=*s;
		if(c>='0' && c<='9')
			v = v*10 + c-'0';
		else
			break;
		s++;
	}
	if(is_negative)
		*value = -v;
	else
		*value = v;
	return s;
}
static const char *skip_string(const char *s) {
	while(*s==' ' || *s=='\t') s++;
	while(*s!=' ' && *s!='\t' && *s!='\0') s++;
	while(*s==' ' || *s=='\t') s++;
	return s;
}


bool ReadLog(const char *log_filename, EventRecord **erpp, size_t *records) {
	FILE *fp = fopen(log_filename,"r");
	if(!fp) return false;
	setvbuf(fp,0,_IOFBF,32768);
	
	EventRecord *e=0;
	size_t n=0;
	size_t a=0;

	char line[128];
	while(fgets(line,(int)sizeof(line),fp)) {
		if(n==a) {
			if(a<1024)
				a = 1024;
			else
				a = a + a/2;
			e = (EventRecord*)realloc(e,sizeof(EventRecord)*a);
			if(!e) {
				fprintf(stderr,"out of memory\n");
				fclose(fp);
				return false;
			}
		}
		unsigned long timestamp_sec;
		unsigned long timestamp_usec;
		//sscanf(line,"%lx %lu.%lu %d %*[^ ] %lx %lx %ld",
		//       &e[n].tid,
		//       &timestamp_sec,
		//       &timestamp_usec,
		//       &e[n].event,
		//       &e[n].mtx_addr,
		//       &e[n].cond_addr,
		//       &e[n].info
		//      );
		const char *s = line;
		s = parse_lx(s,&e[n].tid);
		s = parse_lu(s,&timestamp_sec);
		if(*s=='.') s++;
		s = parse_lu(s,&timestamp_usec);
		s = parse_d(s,&e[n].event);
		s = skip_string(s);
		s = parse_lx(s,&e[n].mtx_addr);
		s = parse_lx(s,&e[n].cond_addr);
		s = parse_ld(s,&e[n].info);
		
		if(e[n].event==ev_thread_name ||
		   e[n].event==ev_mutex_name ||
		   e[n].event==ev_cond_name)
		{
			char name[512];
			sscanf(line,"%lx %lu.%lu %d %*[^ ] %lx %lx %s",
			       &e[n].tid,
			       &timestamp_sec,
			       &timestamp_usec,
			       &e[n].event,
			       &e[n].mtx_addr,
			       &e[n].cond_addr,
			       name
			      );
			size_t namelen=strlen(name);
			char *pname = (char*)malloc(namelen+1);
			memcpy(pname,name,namelen+1);
			e[n].name = pname;
		}
		e[n].timestamp = timestamp_sec + timestamp_usec/1000000.0;
		n++;
		if((n%32768)==0) {
			printf("\nRead %ld records...",(long)n);
			fflush(stdout)	;
		}
	}
	fclose(fp);
	*erpp = e;
	*records = n;
	printf("\nRead %ld records...",(long)n);
	return true;
}
